-------------------------------------------------------------------------------
-- Localization libraries
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")
local BS = AceLibrary("Babble-Spell-2.2")

-------------------------------------------------------------------------------
-- Aliases & constants
-------------------------------------------------------------------------------

local TEH_BUTTON_NAME = "SimpleSelfRebuffButton"

local trackings = {}
do
	local function initTrackings(...)
		for i=1,select('#', ...) do
			local t = select(i, ...)			
			trackings[BS:GetSpellIcon(t)] = BS[t]
		end
	end
	initTrackings(
		'Track Beasts',     'Track Demons', 'Track Dragonkin',
		'Track Elementals', 'Track Giants', 'Track Hidden',
		'Track Humanoids',  'Track Undead', 'Find Herbs',
		'Find Minerals',    'Find Treasure'
	)
end

local activationCheckEvents = {
	'PLAYER_DEAD', 'PLAYER_UNGHOST', 'PLAYER_ALIVE',
	'PLAYER_UPDATE_RESTING',
	'PLAYER_REGEN_DISABLED', 'PLAYER_REGEN_ENABLED',
	'PLAYER_CONTROL_GAINED', 'PLAYER_CONTROL_LOST',
	'SimpleSelfRebuff_CheckActivation'
}

local CATEGORY_TRACKING = L["Tracking"]
local CATEGORY_MAINHAND = L["Main weapon"]
local CATEGORY_OFFHAND  = L["Off-hand weapon"]

local SOURCE_SPELL = 'spell'
local SOURCE_ITEM  = 'item'

local TARGET_AURA     = 'aura'
local TARGET_TRACKING = 'tracking'
local TARGET_MAINHAND = 'main-hand'
local TARGET_OFFHAND  = 'off-hand'

local SLOT_MAINHAND = GetInventorySlotInfo('MainHandSlot')
local SLOT_OFFHAND  = GetInventorySlotInfo('SecondaryHandSlot')

local invSlots = {
	[CATEGORY_MAINHAND] = SLOT_MAINHAND,
	[CATEGORY_OFFHAND]  = SLOT_OFFHAND,
	[TARGET_MAINHAND]   = SLOT_MAINHAND,
	[TARGET_OFFHAND]    = SLOT_OFFHAND,	
}

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local categories = {}
local actualBuffs = {}
local actualBuffExpirations = {}

local buffToCast
local castButton

local active

local db

local sources = {}
local targets = {}

-------------------------------------------------------------------------------
-- Externals
-------------------------------------------------------------------------------

local AceOO = AceLibrary('AceOO-2.0')
local Tablet = AceLibrary('Tablet-2.0')
local Gratuity = AceLibrary('Gratuity-2.0')
 
local Mixin = AceOO.Mixin
local Class = AceOO.Class
local Classpool = AceOO.Classpool

-------------------------------------------------------------------------------
-- Addon declaration
-------------------------------------------------------------------------------

SimpleSelfRebuff = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
local SimpleSelfRebuff = SimpleSelfRebuff

SimpleSelfRebuff.options = {	
	type = 'group',
	args = {
		buffHeader = {
			type = 'header',
			name = ' ',
			order = 199,
		}
	}
}
	
do
	local self = SimpleSelfRebuff
	SSR = self
	
	self.CATEGORY_TRACKING = CATEGORY_TRACKING
	self.CATEGORY_MAINHAND = CATEGORY_MAINHAND 
	self.CATEGORY_OFFHAND  = CATEGORY_OFFHAND 
	
	self.SOURCE_SPELL = SOURCE_SPELL
	self.SOURCE_ITEM  = SOURCE_ITEM
	
	self.TARGET_AURA     = TARGET_AURA
	self.TARGET_TRACKING = TARGET_TRACKING 
	self.TARGET_MAINHAND = TARGET_MAINHAND
	self.TARGET_OFFHAND  = TARGET_OFFHAND

	self.categories = categories
	self.sources = sources
	self.targets = targets
	self.actualBuffs = actualBuffs
	self.actualBuffExpirations = actualBuffExpirations	
end

-------------------------------------------------------------------------------
-- FuBar plugin settings
-------------------------------------------------------------------------------

--SimpleSelfRebuff.hasIcon = true
SimpleSelfRebuff.hideWithoutStandby = true
SimpleSelfRebuff.OnMenuRequest = SimpleSelfRebuff.options

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function error(msg, ...)
	if select('#', ...) > 1 then
		msg = msg:format(...)
	end
	return _G.error(msg)
end

local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		if type(t) == 'table' then
			for k in pairs(t) do
				t[k] = nil 
			end
		end
	end
end

local function colorize(text, r, g, b)
	if type(r) == 'table' and r.r then
		return colorize(msg, r.r, r.g, r.b)
	else
		return ("|cff%02x%02x%02x%s|r"):format(math.floor(r*255), math.floor(g*255), math.floor(b*255), text)
	end
end

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == 'table' then
			setmetatable(t, nil)
			for k in pairs(t) do
				t[k] = nil
			end
			list[t] = true
		end
		return nil
	end
end

-------------------------------------------------------------------------------
-- Initializing
-------------------------------------------------------------------------------

-- Addon setup
function SimpleSelfRebuff:OnInitialize()
	self.revision = tonumber(string.sub("$Revision: 42630 $", 12, -3))
	self.version = "r" .. SimpleSelfRebuff.revision
	self.date = string.sub("$Date: 2007-07-03 13:16:27 -0400 (Tue, 03 Jul 2007) $", 8, 17)

	self:RegisterDB("SimpleSelfRebuffDB")
	self:RegisterDefaults('profile', {
		categories = {},
	})
		
	db = self.db.profile

	self:RegisterChatCommand({"/ssr", "/simpleselfrebuff"}, self.options)
	
	if AceLibrary:HasInstance('AceDebug-2.0') then
		AceLibrary('AceDebug-2.0'):embed(SimpleSelfRebuff)
		SimpleSelfRebuff:SetDebugging(UnitName('player') == 'Adirelle')
	else
		function SimpleSelfRebuff.Debug() end
	end
		
end

function SimpleSelfRebuff:OnEnable(first)
	if first then
		self:ProcessRegistry()
	end

	self:RegisterBucketEvent(activationCheckEvents, 0.1, 'CheckActivation')
	self:RegisterBucketEvent('PLAYER_AURAS_CHANGED', 0.2)
	self:RegisterBucketEvent('SimpleSelfRebuff_UpdateDisplay', 0.1, 'UpdateDisplay')

	self:TriggerEvent('SimpleSelfRebuff_CheckActivation')	
end

function SimpleSelfRebuff:PLAYER_AURAS_CHANGED()
	self:TriggerEvent('SimpleSelfRebuff_CheckActivation')	
	if active then
		self:ScanPlayerBuffs()
	end
end

function SimpleSelfRebuff:AvailableBuffsChanged()

	for target,mixin in pairs(targets) do
		erase(mixin.activeBuffs)
		for buff in pairs(mixin.buffs) do
			if buff:IsActive() then
				mixin.activeBuffs[buff.name] = buff
			end
		end
	end

	self:UpdateOptions()
	self:TriggerEvent('SimpleSelfRebuff_CheckActivation')
end

-------------------------------------------------------------------------------
-- Options handling
-------------------------------------------------------------------------------

do
	local NO_BUFF = L['None']
	
	local function getSelectCategoryBuff(categoryName)
		return db.categories[categoryName] or NO_BUFF
	end
	
	local function getToggleCategoryBuff(categoryName)
		return db.categories[categoryName] == categoryName
	end
	
	local function setCategoryBuff(categoryName, name)
		if name == NO_BUFF then
			name = nil
		elseif name == true then
			name = categoryName
		end
		db.categories[categoryName] = name
		SimpleSelfRebuff:CheckRebuff()
	end
	
	function SimpleSelfRebuff:BuildOptions()
		local opts = self.options.args
		for categoryName, category in pairs(categories) do
			local categoryName, category = categoryName, category
			local option = {
				name = categoryName,
				desc = categoryName,
				passValue = categoryName,
					set = setCategoryBuff,
				order = 200,
			}
			if category.count > 1 then
				option.type = 'text'
				option.get = getSelectCategoryBuff
				option.validate = { NO_BUFF }
			else
				option.type = 'toggle'
				option.get = getToggleCategoryBuff
			end
			opts[categoryName] = option
		end
		self:UpdateOptions()
	end
	
	function SimpleSelfRebuff:UpdateOptions()
		for categoryName, category in pairs(categories) do
			if category.count > 1 then
				local validate = self.options.args[categoryName].validate
				erase(validate)
				for name,buff in pairs(category.buffs) do
					if buff:IsActive() then
						tinsert(validate, buff.name)
					end
				end
				table.sort(validate)				
				tinsert(validate, 1, NO_BUFF)
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Base buff class
-------------------------------------------------------------------------------

local BuffClass, ItemBuff, SpellBuff, AuraBuff, TrackingBuff, MainHandBuff, OffHandBuff

BuffClass = Class()

function BuffClass.prototype:init(name, category, recast, ...)
	self.name = name
	self.category = category
	self.recast = recast
	
	for i = 1,select('#', ...),2 do
		local k,v = select(i, ...)
		if not self[k] then
			self[k] = v
		end
	end

	BuffClass.super.prototype.init(self)
	
	targets[self.target].buffs[self] = true
	sources[self.source].buffs[self] = true
end

function BuffClass.prototype:IsActive()
	return self.found and self.recast and self:IsActivable()
end

function BuffClass.prototype:IsActivable()
	if self.minLevel and UnitLevel('player') < self.minLevel then
		return false
	elseif self.maxLevel and UnitLevel('player') > self.maxLevel then
		return false
	elseif self.powerType and UnitPowerType('player') ~= self.powerType then
		return false
	end
	return true
end

-------------------------------------------------------------------------------
-- Some mixin helpers
-------------------------------------------------------------------------------

local function BuffMixin(...)
	local mixin = Mixin{...}
	mixin.buffs = {}
	return mixin
end

local function BuffSourceMixin(source, ...)
	mixin = BuffMixin("IsUsable", "SetupBuffSource", "source", ...)
	mixin.source = source
	sources[source] = mixin
	return mixin
end

local function BuffTargetMixin(target, ...)
	mixin = BuffMixin("SetupBuffTarget", "target", ...)
	mixin.target = target
	mixin.activeBuffs = {}
	targets[target] = mixin
	return mixin
end

-------------------------------------------------------------------------------
-- Buffs from items
-------------------------------------------------------------------------------

ItemBuff = BuffSourceMixin(SOURCE_ITEM)

function ItemBuff:OnInstanceInit(buff)
	if not buff.itemid then
		error("itemid is missing for ItemBuff %q", buff.name)
	end
end

function ItemBuff:OnEnable()
	if next(self.buffs) then
		SimpleSelfRebuff:RegisterBucketEvent('BAG_UPDATE', 0.1, function() self:ScanItems() end)
		self:ScanItems()		
	end
end

function ItemBuff:OnDisable()
	if SimpleSelfRebuff:IsBucketEventRegistered('BAG_UPDATE') then
		SimpleSelfRebuff:UnregisterBucketEvent('BAG_UPDATE')
	end
end

do
	local function CacheItem(itemsToCache)
		local itemid = itemsToCache and next(itemsToCache)
		if itemid then
			itemsToCache[itemid] = nil
			Gratuity:SetHyperlink("item:"..itemid)
		else
			SimpleSelfRebuff:CancelScheduledEvent('SimpleSelfRebuff_CacheItems')
			itemsToCache = del(itemsToCache)
			ItemBuff:ScanItems()
		end
	end

	function ItemBuff:ScanItems()
		local dirty
		local itemsToCache
	
		for buff in pairs(self.buffs) do
			if buff.itemid then
				local itemName = GetItemInfo(buff.itemid)
				if not itemName then
					if not itemsToCache then
						itemsToCache = new()
					end
					itemsToCache[buff.itemid] = true
				elseif itemName and not buff.found then
					buff.found = true							
					dirty = true			
					if itemName ~= buff.name then
						local category = buff.category
						category.buffs[buff.name] = nil
						buff.name = itemName
						category.buffs[itemName] = buff
					end
				elseif not itemName and buff.found then
					buff.found = nil
					dirty = true						
				end
			end
		end
		
		if itemsToCache then
			SimpleSelfRebuff:ScheduleRepeatingEvent('SimpleSelfRebuff_CacheItems', CacheItem, 0.4, itemsToCache)
		end
		
		if dirty then
			SimpleSelfRebuff:AvailableBuffsChanged()
		end
	end
end

function ItemBuff:IsUsable()
	local start, duration = GetItemCooldown(self.itemid)
	if start > 0 and duration > 0 and start + duration > GetTime() then
		return false
	end					
	return (IsUsableItem(self.itemid))
end

function ItemBuff:SetupBuffSource(castButton)
	SimpleSelfRebuff:Debug('Setup using for item %q', self.name)
	castButton:SetAttribute('type', 'item')
	castButton:SetAttribute('item', self.itemid)
end

-------------------------------------------------------------------------------
-- Buffs from spells
-------------------------------------------------------------------------------

SpellBuff = BuffSourceMixin(SOURCE_SPELL)

function SpellBuff:OnEnable()
	if next(self.buffs) then
		SimpleSelfRebuff:RegisterBucketEvent('SPELLS_CHANGED', 0.1, function() self:ScanSpells() end)
		self:ScanSpells()
	end
end

function SpellBuff:OnDisable()
	if SimpleSelfRebuff:IsBucketEventRegistered('SPELLS_CHANGED') then
		SimpleSelfRebuff:UnregisterBucketEvent('SPELLS_CHANGED')
	end
end

function SpellBuff:ScanSpells()
	local dirty = false
	local knownSpells = new()
	
	local i = 1
	while true do
		local name = GetSpellName(i, BOOKTYPE_SPELL)
		if not name then
			break
		end
		knownSpells[name] = true
		i = i + 1
	end
	
	for buff in pairs(self.buffs) do
		local found = knownSpells[buff.name]
		if found and not buff.found then
			-- SimpleSelfRebuff:Debug('New spell: %q', buff.name)
			buff.found = true
			dirty = true
		elseif not found and buff.found then
			-- SimpleSelfRebuff:Debug('Old spell: %q', buff.name)
			buff.found = nil
			dirty = true
		end
	end
	
	knownSpells = del(knownSpells)

	if dirty then
		SimpleSelfRebuff:AvailableBuffsChanged()
	end	
end

function SpellBuff:IsUsable()
	local start, duration = GetSpellCooldown(self.name)
	if start > 0 and duration > 0 and start + duration > GetTime() then
		return false
	end					
	return (IsUsableSpell(self.name))
end

function SpellBuff:SetupBuffSource(castButton)
	SimpleSelfRebuff:Debug('Setup casting for spell %q', self.name)
	castButton:SetAttribute('type', 'spell')
	castButton:SetAttribute('spell', self.name)
end

-------------------------------------------------------------------------------
-- Auras
-------------------------------------------------------------------------------

AuraBuff = BuffTargetMixin(TARGET_AURA)

function AuraBuff:ScanCurrentBuffs(callback)
	for i = 1, 32 do
		local index = GetPlayerBuff(i, "HELPFUL|PASSIVE")
		local buffName = index and GetPlayerBuffName(index)		
		local buff = buffName and self.activeBuffs[buffName]
		if buff then
			callback(buff, GetPlayerBuffTimeLeft(index))
		end
	end	
end

function AuraBuff:SetupBuffTarget(castButton)
	castButton:SetAttribute('unit', 'player')
end

-------------------------------------------------------------------------------
-- Tracking
-------------------------------------------------------------------------------

TrackingBuff = BuffTargetMixin(TARGET_TRACKING)
TrackingBuff.textures = {}

function TrackingBuff:OnInstanceInit(buff)
	local texture = BS:GetSpellIcon(buff.name)
	if not texture then
		error('Unknown tracking texture for spell %q', buff.name)
	elseif self.textures[texture] then
		error('Tracking texture already set for %q: %q', buff.name, self.textures[texture].name)
	end
	self.textures[texture] = buff
end

function TrackingBuff:ScanCurrentBuffs(callback)
	local texture = GetTrackingTexture()
	local buff = texture and self.textures[texture]
	if buff then
		callback(buff)
	end
end

function TrackingBuff:SetupBuffTarget(castButton)
	castButton:SetAttribute('unit', 'player')
end

-------------------------------------------------------------------------------
-- Weapon enchantment superbuff
-------------------------------------------------------------------------------

do
	local function ScanCurrentBuffs(self, callback)
		local hasEnchant, expiration = select(self.weaponEnchantIndex, GetWeaponEnchantInfo())
		if hasEnchant then
			local strs = new()
			for name, buff in pairs(buffs) do
				if buff.found then
					tinsert(strs, '('..buff.name..')')
				end
			end
			Gratuity:SetInventoryItem(self.slot)
			local _,_,name = Gratuity:MultiFind(unpack(strs))
			if name and buffs[name] then
				seenBuff(buffs[name], expiration/1000)
			end
			strs = del(strs)
		end
	end
	
	local function SetupBuffTarget(self, castButton)
		castButton:SetAttribute('target-slot', self.slot)
	end
	 
	local function WeaponBuffMixin(target, slot, index)
		local mixin = BuffTargetMixin(target, "slot")
		mixin.slot = GetInventorySlotInfo(slot)
		mixin.weaponEnchantIndex = index
		mixin.ScanCurrentBuffs = ScanCurrentBuffs
		mixin.SetupBuffTarget = SetupBuffTarget
		return mixin
	end
	
	MainHandBuff = WeaponBuffMixin(TARGET_MAINHAND, 'MainHandSlot', 1)
	OffHandBuff = WeaponBuffMixin(TARGET_OFFHAND, 'SecondaryHandSlot', 3)
	
end

-------------------------------------------------------------------------------
-- Buff registry
-------------------------------------------------------------------------------

do
	local buffFuncs = {}

	function SimpleSelfRebuff:RegisterBuffSetup(func)
		if type(func) == "function" then
			tinsert(buffFuncs, func)
		else
			error("Arg #1 to RegisterBuffs must be a function")
		end
	end
	
	function SimpleSelfRebuff:ProcessRegistry()
		if not buffFuncs then
			return
		end
			
		local selfClassName = select(2, UnitClass('player'))
		for i,func in ipairs(buffFuncs) do
			func(self, BS, selfClassName, L)
		end
		
		buffFuncs = del(buffFuncs)

		self:BuildOptions()
	end
	
	function SimpleSelfRebuff:RegisterCategory(name, ...)
		local category = categories[name]
		if not category then
			category = {
				name = name,
				count = 0,
				buffs = {},
			}
			if select('#', ...) > 0 then
				for i = 1, select('#', ...), 2 do
					local k,v = select(i, ...)
					category[k] = v
				end
			end	
			categories[name] = category
		end
		return category
	end
	
	function SimpleSelfRebuff:RegisterBuff(name, categoryName, recast, target, source, ...)
	
		categoryName = categoryName or name
		local category = self:RegisterCategory(categoryName)
		source = source or SOURCE_SPELL
		target = target or TARGET_AURA
		recast = (recast == nil or recast) and true
		if category.buffs[name] then
			error("%q already registered in category", name)
		end
		if not sources[source] then
			error("unknown source for %q: %q", name, source)
		end		
		if not targets[target] then
			error("unknown target for %q: %q", name, target)
		end		
			
		local class = Classpool(BuffClass, sources[source], targets[target])
		local buff = class:new(name, category, recast, ...)

		category.buffs[name] = buff
		category.count = category.count + 1
		
		return buff
	end
	
end

-------------------------------------------------------------------------------
-- Activation
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckActivation()
	local shouldActivate = not (UnitIsDeadOrGhost('player') or InCombatLockdown() or IsResting()
	  or IsMounted() or IsFlying() or IsStealthed())

	if shouldActivate and not active then
		active = true
		self:Debug('Activated')
		
		for name,source in pairs(sources) do
			source:OnEnable()
		end		
	
		self:ScheduleRepeatingEvent('SimpleSelfRebuff_CheckRebuff', self.CheckRebuff, 1.5, self)
		self:ScanPlayerBuffs()
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
		
	elseif not shouldActivate and active then
		active = nil
		self:Debug('Disactivated')
		
		erase(actualBuffs, actualBuffExpirations)

		self:CleanupPendingCast()		
		self:CancelScheduledEvent('SimpleSelfRebuff_CheckRebuff')

		for name,source in pairs(sources) do
			source:OnDisable()
		end		
		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

-------------------------------------------------------------------------------
-- Core
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckRebuff()

	local newBuffToCast = nil	
	local shortestBuff, shortestBuffDuration
	
	for categoryName,category in pairs(categories) do
		local expectedBuff = category.buffs[db.categories[categoryName] or "-"]		
		if expectedBuff and expectedBuff:IsUsable() then			
			local actualBuff = actualBuffs[category] 
			
			-- Check if it is missing
			if not actualBuff then
				newBuffToCast = expectedBuff
				--self:Debug('Missing buff: %q', expectedBuff)
				break
				
			-- Check remaning time
			elseif actualBuff.name == expectedBuff.name and actualBuffExpirations[categoryName] then
				local timeLeft = actualBuffExpirations[categoryName] - GetTime()
				if timeLeft > 0 then
					if not shortestBuffDuration or timeLeft < shortestBuffDuration then
						shortestBuffDuration, shortestBuff = timeLeft, actualBuff
					end
				end
			end
				
		end
	end
	
	if not newBuffToCast and shortestBuffName and shortestBuffDuration < 30 then
		newBuffToCast = shortestBuff
		self:Debug('Buff needs refresh: %q (%q)', newBuffToCast, shortestBuffDuration)
	end

	if newBuffToCast ~= buffToCast then
		
		if newBuffToCast then
			buffToCast = newBuffToCast
			
			if not castButton then
				castButton = CreateFrame("Button", TEH_BUTTON_NAME, UIParent, "SecureActionButtonTemplate")
				castButton:RegisterForClicks("LeftButtonUp")
				castButton:SetPoint("CENTER",0,0)
				castButton:Hide()
			end
			
			-- Reset attributes
			castButton:SetAttribute('type', nil)
			castButton:SetAttribute('spell', nil)			
			castButton:SetAttribute('unit', nil)			
			castButton:SetAttribute('target-slot', nil)			
			
			-- Setup attributes depending on source and target
			--self:PrintLiteral(buffToCast)
			buffToCast:SetupBuffSource(castButton)
			buffToCast:SetupBuffTarget(castButton)
			
			-- Binding
			self:Debug('Binding mousewheel')
			SetOverrideBindingClick(castButton, true, "MOUSEWHEELUP", TEH_BUTTON_NAME, "LeftButton")
			SetOverrideBindingClick(castButton, true, "MOUSEWHEELDOWN", TEH_BUTTON_NAME, "LeftButton")	

		else
			self:CleanupPendingCast()
		end
		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end

end

function SimpleSelfRebuff:CleanupPendingCast()
	if buffToCast then
		self:Debug('Unbinding mousewheel')
		buffToCast = nil
		if castButton then
			ClearOverrideBindings(castButton)
		end
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

-------------------------------------------------------------------------------
-- Player buff status tracking
-------------------------------------------------------------------------------

do
	local seen
	local dirty
	
	local function seenBuff(buff, timeLeft)
		--SimpleSelfRebuff:Debug('Seen %s %q (%q)', buff.target, buff.name, timeLeft)
		local category = buff.category
		seen[category] = true
		if not actualBuffs[category] then
			actualBuffs[category] = buff
			SimpleSelfRebuff:Debug('New %s buff: %q', buff.target, buff.name)			
			dirty = true
		end
		if timeLeft and math.floor(timeLeft) > 0 then
			actualBuffExpirations[category] = GetTime() + timeLeft		
		else
			actualBuffExpirations[category] = nil
		end		
	end
	
	function SimpleSelfRebuff:ScanPlayerBuffs()
		seen = new()
		dirty = nil
		
		for target, mixin in pairs(targets) do
			if next(mixin.activeBuffs) then
				-- self:Debug('Scanning %s', mixin.target)
				mixin:ScanCurrentBuffs(seenBuff)
			end
		end
		
		for category, actualBuff in pairs(actualBuffs) do
			if not seen[category] then
				actualBuffs[category] = nil
				actualBuffExpirations[category] = nil
				self:Debug('Faded out: %q', actualBuff.name)
				dirty = true
			end
		end
		
		if dirty then
			dirty = nil
			self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
			self:Debug('Auras changed - checking rebuff')
			self:CheckRebuff()
		end
		
		seen = del(seen)
		
	end
	
end

-------------------------------------------------------------------------------
-- FuBar Plugin info
-------------------------------------------------------------------------------

local function GetBuffInfo(category)

	local expectedBuff = category.buffs[db.categories[category.name] or "-"]
	if not expectedBuff then
		return '-', '-', 0.5, 0.5, 0.5
	end
	
	local actualBuff = actualBuffs[category]
	if not actualBuff then
		return expectedBuff.name, L['Missing'], 1, 0, 0
	end
	
	local actualExpiration = actualBuffExpirations[category.name]
	local timeLeft
	if actualExpiration then
		timeLeft = math.floor(actualExpiration - GetTime())
		if timeLeft >= 3600 then
			timeLeft = HOUR_ONELETTER_ABBR:format(math.ceil(timeLeft/3600))
		elseif timeLeft >= 60 then
			timeLeft = MINUTE_ONELETTER_ABBR:format(math.ceil(timeLeft/60))
		else
			timeLeft = SECOND_ONELETTER_ABBR:format(timeLeft)
		end	
	else
		timeLeft = L['n/a']
	end
	local color
	if actualBuff.name == expectedBuff.name then
		return actualBuff.name, timeLeft, 0, 1, 0
	else
		return actualBuff.name, timeLeft, 1, 1, 0
	end
end

function SimpleSelfRebuff:OnTextUpdate()
	if active then
		if buffToCast then
			local name, timeLeft, r, g, b = GetBuffInfo(buffToCast.category)
			self:SetText(colorize(timeLeft, r, g, b))
			SimpleSelfRebuff:ScheduleEvent('SimpleSelfRebuff_UpdateDisplay', 1)
			--self:SetIcon(BS:GetSpellIcon(buffToCast))
		else
			self:SetText('SSR')
			--self:SetIcon(false)
		end
	else
		self:SetText('|cff777777SSR|r')
		--self:SetIcon(false)
	end
end

function SimpleSelfRebuff:OnTooltipUpdate()
	local cat
	Tablet:SetTitle(self:GetTitle())
	if not active then
		cat = Tablet:AddCategory()
		cat:AddLine('text', L['Currently inactive (stealthed/in combat/dead/mounted/resting).'])
	else
		cat = Tablet:AddCategory("columns", 3)
		for categoryName, category in pairs(categories) do
			if category.count > 0 then
				local name, timeLeft, r, g, b = GetBuffInfo(category)
				cat:AddLine(
					'text',  categoryName,
					'text2', name,     'text2R', r, 'text2G', g, 'text2B', b,
					'text3', timeLeft, 'text3R', r, 'text3G', g, 'text3B', b
				)
			end
		end
		SimpleSelfRebuff:ScheduleEvent('SimpleSelfRebuff_UpdateDisplay', 1)		
		Tablet:SetHint(L['Green means "ok", yellow means "override wanted buff", red means "missing".'])
	end
end


